/* BBCIM: Acorn BBC computer diskimage manipulation
 *
 * archive_read_attributes.c
 * ENGLISH: reads file attributes from the info file
 *
   Copyright (C) W.H. Scholten 2001
 
   Permission to use, copy, modify, distribute, and sell this software
   and its documentation for any purpose is hereby granted without fee,
   provided that the above copyright notice appear in all copies and
   that both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of the copyright holder
   not be used in advertising or publicity pertaining to distribution
   of the software without specific, written prior permission. The
   copyright holder makes no representations about the suitability of
   this software for any purpose. It is provided "as is" without express
   or implied warranty.
  
   THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
   SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
   FITNESS, IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
   SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
   RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
   CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "archive_read_attributes.h"


/* Returns 0 if ok, anything else if not */
int bbcim_a_r_a_string_to_hex(char *string, long *hex) {
	unsigned int  i = 0;
	int  st;

	for (i=0; i<strlen(string); i++) {
		if (string [ i ] != ' ') break;
	}

	st = 0;
	for (   ; i<strlen(string); i++) {
		if (!isxdigit(string [ i ])) { st = -1; break; }
	}
	if (st == 0) st = sscanf(string, "%lx", hex) - 1;

	return st;
}


int bbcim_archive_read_file_attributes(char *filename, struct bbcfile *file, int verbose) {
	char *info_arg[] = {
		"TAPE", "filename", "load address", "exec address", "length", "locked", "CRC=", "crc", "NEXT", "next tapefile"
	};
	int  info_optional[] = { 1, 0, 0, 0, 1, 1, 0, 0, 1, 1 };
//	int  info_optional[] = { 1, 0, 0, 0, 1, 1, 1, 1, 1, 1 }; /* CRC */
	int  st;
	char *filename_inf = NULL;
	FILE *fpinfo = NULL;
	char string[64];
	int  item_wanted;
	int  status = -1; /* function's return status */

	file->name = NULL;
	file->next_tape_file = NULL;

	filename_inf = (char *) malloc(strlen(filename+1+4));
	if (filename_inf == NULL) {
		fprintf(stderr, "Out of memory\n");
		return -1;
	}

	strcpy(filename_inf, filename);
	strcat(filename_inf, ".inf");

	fpinfo = fopen(filename_inf, "r");
	if (fpinfo == NULL) {
		if (verbose) printf("Unable to open %s\n", filename_inf);
		goto leave_read_file_attributes;
	}

#define BBC_ARCHIVE_ATTR_TAPE          0
#define BBC_ARCHIVE_ATTR_NAME          1
#define BBC_ARCHIVE_ATTR_LOADADDR      2
#define BBC_ARCHIVE_ATTR_EXECADDR      3
#define BBC_ARCHIVE_ATTR_LENGTH        4
#define BBC_ARCHIVE_ATTR_LOCKED        5
#define BBC_ARCHIVE_ATTR_CRCEQUALS     6
#define BBC_ARCHIVE_ATTR_CRC           7
#define BBC_ARCHIVE_ATTR_NEXT          8
#define BBC_ARCHIVE_ATTR_NEXTTAPEFILE  9

	item_wanted = BBC_ARCHIVE_ATTR_TAPE;
	for (;;) {
		int  j;

		/* Direct string reading as e.g. the filename may contain control
		 * characters */
		j = 0;
		string[0] = 0;
		for (;;) {
			st = fread(string, 1, 1, fpinfo);
			//printf("AA st=%d, read %c\n", st, string[0]);
			if (feof(fpinfo)) break;
			if (string [ j ] != ' ') { /* Found the first char */
				for (j=1; j<63; j++) {
					st = fread(string + j, 1, 1, fpinfo);
					//printf("BB st=%d, read %c\n", st, string[j]);
					if ((string [ j ] == 13 || string [ j ] == 10) && item_wanted >= BBC_ARCHIVE_ATTR_LOADADDR) break; /* nuke tailing CR/LF */
					if (string [ j ] == ' ') break;
					if (feof(fpinfo)) break;
				}
				string [ j ] = 0;
				if (j == 63) {
					if (verbose) fprintf(stderr, "%s: string too long (partial = %s)\n", filename_inf, string);
					goto leave_read_file_attributes;
				}
				break;
			}
		}
		st = strlen(string);



		if (st <= 0) {
			if (item_wanted >= BBC_ARCHIVE_ATTR_CRC + 1)
				break;
			else {
				while (info_optional[ item_wanted ]) item_wanted ++;

				if (verbose) fprintf(stderr, "%s: missing argument (%s)\n", filename_inf, info_arg[ item_wanted ]);
				goto leave_read_file_attributes;
			}
		}

		switch (item_wanted) {
		case BBC_ARCHIVE_ATTR_TAPE:
			if (!strcmp(string, "TAPE")) {
				item_wanted ++;
				file->type = BBCIM_FILETYPE__TAPE;
				continue;
			}

			/* Disk file */
			file->type = BBCIM_FILETYPE__DISK;
			/* Fall through .. */
			item_wanted ++;

		case BBC_ARCHIVE_ATTR_NAME:
			if (file->type == BBCIM_FILETYPE__DISK) {
				file->name = strdup(string);
			}
			if (file->type == BBCIM_FILETYPE__TAPE) {
				int  i;

				file->name = (char *) malloc(strlen(string)+1+2);

				if (strlen(string) > 10) {
					if (verbose) fprintf(stderr, "%s: tape name %s: too long\n", filename_inf, string);
					goto leave_read_file_attributes;
				}

				/* Tape filenames converted to $.<tapename> for uniform
				 * handling of tape/disk. */
				file->name[0] = '$';
				file->name[1] = '.';
				for (i=strlen(string); i>=0; i--) file->name[i+2] = string[i];
			}

			break;

		case BBC_ARCHIVE_ATTR_LOADADDR:
			st = bbcim_a_r_a_string_to_hex(string, &file->loadaddr);

			if (st != 0) {
				if (verbose) fprintf(stderr, "%s: read load address failed\n", filename_inf);
				goto leave_read_file_attributes;
			}
			break;

		case BBC_ARCHIVE_ATTR_EXECADDR:
			st = bbcim_a_r_a_string_to_hex(string, &file->execaddr);
			if (st != 0) {
				if (verbose) fprintf(stderr, "%s: read exec address failed\n", filename_inf);
				goto leave_read_file_attributes;
			}
			break;

		case BBC_ARCHIVE_ATTR_LENGTH:
			st = bbcim_a_r_a_string_to_hex(string, &file->length);
			if (st == 0) {
				break;
			}
			/* Fall through .. */
			item_wanted ++;

		case BBC_ARCHIVE_ATTR_LOCKED:
			file->attributes = 0;
			if ((!strcmp(string, "L")) || (!strcmp(string, "Locked"))) {
				file->attributes |= BBCIM_ATTR__LOCKED;
				break;
			}

			/* Fall through .. */
			item_wanted ++;

		case BBC_ARCHIVE_ATTR_CRCEQUALS:
			if (strlen (string) >= 4 && !strncmp(string, "CRC=", 4)) {
				fseek(fpinfo, -(strlen(string)-4), SEEK_CUR);
			}
			else {
				/* Crc not found */
				file->crc = -1;
				/* should be a fatal error */
				//goto leave_read_file_attributes;
				item_wanted ++; /* skip BBC_ARCHIVE_ATTR_CRC */
			}
			break;

		case BBC_ARCHIVE_ATTR_CRC:
			st = bbcim_a_r_a_string_to_hex(string, &file->crc);
			if (st != 0) {
				file->crc = -1;
				/* should be a fatal error */
				//goto leave_read_file_attributes;
			}
			else {
				/* Check the range */
				if (file->crc < 0 || file->crc > 65535) {
					file->crc = -1;
				}
			}
			break;

		case BBC_ARCHIVE_ATTR_NEXT:
			if (strcmp(string, "NEXT")) item_wanted ++; /* no NEXT so skip NEXTTAPEFILE */
			break;
		case BBC_ARCHIVE_ATTR_NEXTTAPEFILE:
			file->next_tape_file=strdup(string);
			break;
		}
		item_wanted ++;

	}

	status = 0;

  leave_read_file_attributes:
	if (filename_inf) free(filename_inf);
	if (fpinfo) fclose(fpinfo);

	return status;
}


void bbcim_archive_free_file_attributes(struct bbcfile *file) {
	if (file->name)
		free(file->name);

	if (file->next_tape_file)
		free(file->next_tape_file);
}



int bbcim_diskinfo_from_string(char *string, struct bbcdiskimage *image) {
	if (!strncmp(string, "DFS", 10)) image->type = BBCIM_DFS_DISK;
	else if (!strncmp(string, "W62", 10)) image->type = BBCIM_W62_DISK;
	else if (!strncmp(string, "DDOS", 10)) image->type = BBCIM_DDOS_DISK;
	else if (!strncmp(string, "ADFS", 10)) image->type = BBCIM_ADFS_DISK;
	else if (!strncmp(string, "HDFS", 10)) image->type = BBCIM_HDFS_DISK;
	else if (!strncmp(string, "DS", 10)) image->sides = 2;
	else if (!strncmp(string, "SS", 10)) image->sides = 1;
	else return -1;

	return 0;
}

int bbcim_archive_read_disk_attributes(char *diskname, struct bbcdiskimage *image, int verbose) {
	FILE *fpdsk;
	char *dsk;
	char type [ 11 ];
	char sides [ 11 ];
	int  status = -1;

	dsk = (char *) malloc(strlen(diskname)+1+4);
	if (dsk == NULL) {
		fprintf(stderr, "Out of memory\n");
		return -1;
	}
	strcpy(dsk, diskname);
	strcat(dsk, ".dsk");

	fpdsk = fopen(dsk, "r");
	if (fpdsk == NULL) {
		if (verbose) fprintf(stderr, "Unable to open %s\n", dsk);
		return -1;
	}

	/* Read disk attributes */
	if (fscanf(fpdsk, "%10s %10s %dT %dS %dB", type, sides, &image->tracks, &image->sectors_per_track, &image->bytes_per_sector) != 5) {
		if (verbose) fprintf(stderr, "Bad .dsk format\n");
		goto leave_read_disk_attributes;
	}

	if (bbcim_diskinfo_from_string(type, image) || bbcim_diskinfo_from_string(sides, image)) {
		if (verbose) fprintf(stderr, "Bad .dsk format\n");
		goto leave_read_disk_attributes;
	}
	status = 0;

  leave_read_disk_attributes:
	if (dsk) free(dsk);
	if (fpdsk) fclose(fpdsk);

	return status;
}


void bbcim_free_disk_attributes(struct bbcdiskimage *disk) {
}
