/*
    make_tape_blocks, by W.H.Scholten, 2001,
    based on 'MakeUEF' by Thomas Harte, 2000

	Distributed under the GPL - see 'Copying' for details

*/

#include <stdlib.h>
#include <ctype.h>
#include <sys/stat.h>
#include <string.h>
#include "defines.h"

#define PROGRAM_NAME	"make_tape_blocks"


#define ORIGIN_BLOCK	0x0000
#define LEADIN_BLOCK	0x0110
#define DATA_BLOCK	0x0100
#define GAP_BLOCK	0x0112

char *inname, *outname, *error, *bstring;

unsigned char buffer[131072], *cpos;

int delays[20];
unsigned char cbyte, sparebits;
bool h2400, bit, bit_error, byte_error;

int understood_one, understood_ten;

bool sign;
int span;

enum span_calibration_types
{
	FREQUENCY_CALIBRATE, WAVE_CALIBRATE, VALUE_CALIBRATE
} span_type;

enum leadin_tone_types
{
	DUMMYBIT_TONE, LONG_TONE
} leadin_type;

C_snd_input *i_drv;
C_raw raw_drv;
C_wav wav_drv;
C_sndcard card_drv;


void (* get_next_length)(void);
void get_next_length_sign();
void get_next_length_peak();


#include "_bbcim_code.c"
#include "_parse_line.cpp"

void push_along_delays(void)
{
	int c;
	c = 19;
	while(c--)
	{
		delays[c+1] = delays[c];
	}
}

void get_next_length_sign(void)
{
	push_along_delays();
	delays[0] = 0;

#if 1
	int ct;
	for (;;) {
		int val = i_drv->GetNextValue();
		delays[0]++;
		if (val > 0 || i_drv->Finished()) break;
	}
	for (;;) {
		int val = i_drv->GetNextValue();
		delays[0]++;
		if (val < 0 || i_drv->Finished()) break;
	}

#else
	do
	{
		sign = i_drv->GetNextValue() < 0;
		delays[0]++;
	}
	while(sign && !i_drv->Finished());

	do
	{
		sign = i_drv->GetNextValue() < 0;
		delays[0]++;
	}
	while(!sign && !i_drv->Finished());
#endif
}

void get_next_length_peak(void)
{
	char value, old_value;

	push_along_delays();
	delays[0] = 1;

	value = i_drv->GetNextValue();

	do
	{
		old_value = value;
		value = i_drv->GetNextValue();
		delays[0]++;
	}
	while(value>=old_value && !i_drv->Finished());

	do
	{
		old_value = value;
		value = i_drv->GetNextValue();
		delays[0]++;
	}
	while(value<=old_value && !i_drv->Finished());
}


void advance(int len)
{
	while(len--)
		get_next_length();
}

void calibrate_span(void)
{
	switch(span_type)
	{
		case FREQUENCY_CALIBRATE :
			span = i_drv->GetFrequency() /1200;
		break;

		case WAVE_CALIBRATE :
			advance(2);
			span = delays[0] + delays[1];
		break;
	}
}

int get_bit(int spos)
{
	if(delays[spos] > (span << 1))	//a very wide margin for error
		bit_error = true;

	if(abs(delays[spos] - (span >> 1)) > 2)
	{
		//low/high
		spos--;

		bit = false;
	}
	else
	{
		//low/high/low/high
		spos-=2;

		bit = true;
	}

	return spos;
}

void re_eval_byte(void)
{
	int bufpos = 19;
	int count;
	sparebits = cbyte = 0;

	bufpos = get_bit(bufpos);
	understood_one = 19-bufpos;
	if(bit)
		sparebits |= 1;

	count = 8;
	while(count--)
	{
		cbyte >>= 1;

		bufpos = get_bit(bufpos);
		cbyte |= bit ? 0x80 : 0;
	}

	bufpos = get_bit(bufpos);
	understood_ten = 19-bufpos;

	if(bit)
		sparebits |= 2;

	byte_error = bit_error || !(sparebits&2);
	h2400 = ((cbyte == 0xff) && (sparebits == 3)) ? true : false;
}

void init_state(void)
{
	advance(20);
	re_eval_byte();
}

void get_block(void)
{
	cpos = buffer;
	memset(buffer, 0, 131072);

	while(!h2400 && !i_drv->Finished())
	{
		advance(1);
		re_eval_byte();
	}

	while(h2400 && !i_drv->Finished())
	{
		advance(1);
		re_eval_byte();
	}

	do
	{
		advance(understood_one);
		re_eval_byte();
	}
	while((cbyte != 0x2a) && !i_drv->Finished());

	byte_error = bit_error = false;
	while(!h2400 && !i_drv->Finished() && !byte_error)
	{
		*cpos++ = cbyte;
		advance(understood_ten);
		re_eval_byte();
	}
}

unsigned char h, l;

void get_crc(unsigned char *start, int len)
{
	unsigned char x;

	h = l = 0;
	while(len--)
	{
		h ^= *start;

		x = 8;
		while(x--)
		{
			if(h&128)
			{
				h ^= 8;
				l ^= 16;

				h = (h << 1) | (l >> 7);
				l = (l << 1) | 1;
			}
			else
			{
				h = (h << 1) | (l >> 7);
				l = (l << 1);
			}
		}

		start++;
	}
}

bool check_block(void)
{
	int eon;
	int length;

	if(byte_error)
		return true;

	eon = strlen((char *)buffer+1)+2;

	if(cpos-buffer < eon)
		return true;

	get_crc(buffer+1, eon+16);

	if(h != buffer[eon+17] || l != buffer[eon+18])
		return true;

	if(!(buffer[eon+12]&0x40))
	{
		length = buffer[eon+10] | (buffer[eon+11] << 8);
		get_crc(buffer+eon+19, length);

		if(h != buffer[eon+19+length] || l != buffer[eon+19+length+1])
			return true;
	}

	return false;
}

void safe_printname(char *name)
{
	char *newname, *namestart;

	namestart = newname = strdup(name);
	while(*namestart)
	{
		if(*namestart < 32 || *namestart > 126)
			*namestart = '?';
		namestart++;
	}

	printf("%-10s", newname);
}

void parse(void)
{
	char *cname;
	int blocknum, newblock, eon;
	bool loading, new_file, first_block = true;
	unsigned char length[2];

	cname = strdup("(new file)");
	new_file = true;

	while(!i_drv->Finished())
	{
		get_block();

		eon = 2 + strlen((char *)buffer + 1);

		/* WHS: block writing */
		if (!check_block()) {
			FILE *fp;;
			char name[1024];
			int  block_no;
			char *block_filename = (char *) (buffer+1);

			strcpy(name, (char *) buffer+1);
			//bbcim_printable_bbcfilename(name);
			bbcim_replace_bad_chars(name);

			/* append "_00" for block no 0, etc */
			unsigned char *d2 = buffer + 1 + strlen(block_filename) + 1;
			d2 += 4;
			d2 += 4;
			block_no = (*d2) | (*(d2+1) << 8);
			sprintf(name+strlen(name), "_%02X", block_no);

//			printf("block %s (", name);
//			safe_printname((char *) (buffer+1));
//			printf(" %02X)", block_no);

			int ii;
			for (ii=0; ; ii++) {
				fp = fopen(name, "r");
				if (fp == NULL) break;

				if (fp != NULL) {
#if 1
					printf("block %s (", name);
					safe_printname((char *) (buffer+1));
					printf(" %02X)", block_no);
					printf(" exists\n");
#endif
					/* Is it the same block? Then skip, otherwise it's from a
					 * different file with the same (host) name, so rename. */
					fseek(fp, 0, SEEK_END);
					int same_block = false;
					if (ftell(fp) == (cpos-buffer)) {
						fseek(fp, 0, SEEK_SET);
						unsigned char block_buf [ 8192 ];
						fread(block_buf, 1, cpos-buffer, fp);
						if (!memcmp(buffer, block_buf, cpos-buffer)) {
							same_block = true;
							printf(" -> Identical to existing block %s, so skipped\n", name);
							break;
						}
					}

					printf("\n -> block differs from %s\n", name);
					fclose(fp);
					fp = NULL;
					/* Next file to try: */
					sprintf(name+strlen(name), "_%d_%02X", ii, block_no);
				}
			}

			if (fp == NULL) { /* Do not overwrite good block files */
				fp = fopen(name, "w");
				if (fp == NULL) {
					perror("write error\n");
					exit(1);
				}
				fwrite(buffer, 1, cpos-buffer, fp);
				printf("Writing: %s", name);
			}
			else { printf("not overwriting\n"); }
			fclose(fp);

			if(buffer[ eon+12 ] & 0x80) {
				printf(" Last block\n");
			}
		}
		else {
			printf("Bad block: (");
			safe_printname((char *) (buffer+1));
			char *block_filename = (char *) (buffer+1);
			unsigned char *d2 = buffer + 1 + strlen(block_filename) + 1;
			d2 += 4;
			d2 += 4;
			int block_no = (*d2) | (*(d2+1) << 8);

			printf(" %02X)", block_no);
		}

		printf("\n");
	}
}

int main(int argc, char **argv)
{
	char result;
	bool print_error = false;

	if(result = parse_line(argv, argc))
 		return result;

	if(i_drv->Open(inname))
	{
		init_state();
		calibrate_span();

		parse();

		i_drv->Close();
	}
	else
		print_error = true;

	if(print_error)
	{
		if(error)
			fprintf(stderr, "ERROR : %s\n", error);
		else
			fprintf(stderr, "Undetermined error\n", error);
		return 3;
	}

	return 0;
}
